@@ -1,5 +1,13 @@
 #!/usr/bin/python3
 
+# Copyright (c) 2017-2023 California Institute of Technology ("Caltech"). U.S.
+# Government sponsorship acknowledged. All rights reserved.
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# You may obtain a copy of the License at
+#
+#     http://www.apache.org/licenses/LICENSE-2.0
+
 '''Routines to (un)project points using any camera model
 
 Most of these are Python wrappers around the written-in-C Python extension
@@ -392,12 +400,13 @@ if get_gradients: we return a tuple:
 
 
     # No projection gradients implemented in C. We should get here approximately
-    # never. At this time, the only projection function that has no gradients
-    # implemented is LENSMODEL_CAHVORE, which nobody is really expected to be
-    # using. If these see use, real gradients should be implemented
+    # never. At this time, there are no longer any projection functions that
+    # have no gradients implemented. If another such model is defined, this path
+    # will be used. If these see use, real gradients should be implemented
     #
     # We compute the gradients numerically. This is a reimplementation of the C
     # code. It's barely maintained, and here for legacy compatibility only
+    raise("should never get here")
 
     if get_gradients:
         raise Exception(f"unproject(..., get_gradients=True) is unsupported for models with no gradients, such as '{lensmodel}'")
