@@ -1,5 +1,13 @@
 #!/usr/bin/python3
 
+# Copyright (c) 2017-2023 California Institute of Technology ("Caltech"). U.S.
+# Government sponsorship acknowledged. All rights reserved.
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# You may obtain a copy of the License at
+#
+#     http://www.apache.org/licenses/LICENSE-2.0
+
 '''General utility functions used throughout mrcal
 
 All functions are exported into the mrcal module. So you can call these via
@@ -453,7 +461,7 @@ RETURNED VALUE
 
     observations_board = optimization_inputs.get('observations_board')
     if observations_board is None:
-        return Exception("No board observations available")
+        raise Exception("No board observations available")
 
     indices_frame_camintrinsics_camextrinsics = \
         optimization_inputs['indices_frame_camintrinsics_camextrinsics']
@@ -1021,7 +1029,7 @@ A list of 'set' directives passable as plot options to gnuplotlib
     except: pass
     try:    imeas0.append(mrcal.measurement_index_points        (0, **optimization_inputs))
     except: pass
-    try:    imeas0.append(mrcal.measurement_index_regularization(0, **optimization_inputs))
+    try:    imeas0.append(mrcal.measurement_index_regularization(**optimization_inputs))
     except: pass
 
     return [f"arrow nohead from {x},graph 0 to {x},graph 1" for x in imeas0]
@@ -1192,8 +1200,13 @@ def _plot_args_points_and_covariance_ellipse(q, what):
 
 def residuals_chessboard(optimization_inputs,
                          *,
-                         i_cam     = None,
-                         residuals = None):
+                         icam_intrinsics     = None,
+                         residuals           = None,
+                         return_observations = False,
+
+                         # for backwards-compatibility only; same as
+                         # icam_intrinsics
+                         i_cam               = None):
     r'''Compute and return the chessboard residuals
 
 SYNOPSIS
@@ -1202,16 +1215,17 @@ SYNOPSIS
 
     gp.plot( mrcal.residuals_chessboard(
                optimization_inputs = model.optimization_inputs(),
-               i_cam               = i_cam ).ravel(),
+               icam_intrinsics     = icam_intrinsics ).ravel(),
              histogram = True,
              binwidth  = 0.02 )
 
     ... A plot pops up showing the empirical distribution of chessboard fit
-    ... errors in this solve. For ALL the cameras
+    ... errors in this solve. For the given camera only
 
 Given a calibration solve, returns the residuals of chessboard observations,
 throwing out outliers and, optionally, selecting the residuals from a specific
-camera.
+camera. These are the weighted reprojection errors present in the measurement
+vector.
 
 ARGUMENTS
 
@@ -1219,48 +1233,216 @@ ARGUMENTS
   from mrcal.optimize(). This describes the solved optimization problem that
   we're visualizing
 
-- i_cam: optional integer to select the camera whose residuals we're visualizing
-  If omitted or None, we report the residuals for ALL the cameras together.
+- icam_intrinsics: optional integer to select the camera whose residuals we're
+  visualizing If omitted or None, we report the residuals for ALL the cameras
+  together.
 
 - residuals: optional numpy array of shape (Nmeasurements,) containing the
   optimization residuals. If omitted or None, this will be recomputed. To use a
   cached value, pass the result of mrcal.optimize(**optimization_inputs)['x'] or
   mrcal.optimizer_callback(**optimization_inputs)[1]
 
+- return_observations: optional boolean, defaulting to False. if
+  return_observations: we return a tuple (residuals,observations) instead of
+  just residuals
+
+If no chessboards are present in the solve I return arrays of appropriate shape
+with N = 0
+
 RETURNED VALUES
 
-Numpy array of shape (N,2) of all the residuals. N is the number of pixel
-observations remaining after outliers and other cameras are thrown out
+if not return_observations:
+
+  we return a numpy array of shape (N,2) of all the residuals. N is the number
+  of pixel observations remaining after outliers and other cameras are thrown
+  out
+
+else:
+
+  we return a tuple:
+
+  - The same residuals array as before
+
+  - The corresponding observation points in a numpy array of shape (N,2). These
+    are a slice of observations_board[] corresponding to each residual
 
     '''
 
+    # shape (Nobservations, object_height_n, object_width_n, 3)
+    observations_board = optimization_inputs.get('observations_board')
+    if observations_board is None or observations_board.size == 0:
+        # no board observations
+        if not return_observations:
+            # shape (N,2)
+            return \
+                np.zeros((0,2), dtype=float)
+        else:
+            # shape (N,2), (N,2)
+            return \
+                np.zeros((0,2), dtype=float), \
+                np.zeros((0,2), dtype=float)
+
+
+
+    # for backwards compatibility
+    if i_cam is not None:
+        icam_intrinsics = i_cam
+
     if residuals is None:
-        # Flattened residuals. The board measurements are at the start of the
-        # array
+        # Flattened residuals. This is ALL the measurements: chessboard, point,
+        # regularization...
         residuals = \
             mrcal.optimizer_callback(**optimization_inputs,
                                      no_jacobian      = True,
                                      no_factorization = True)[1]
 
-    # shape (Nobservations, object_height_n, object_width_n, 3)
-    observations = optimization_inputs['observations_board']
-    residuals_shape = observations.shape[:-1] + (2,)
+    residuals_shape = observations_board.shape[:-1] + (2,)
 
     # shape (Nobservations, object_height_n, object_width_n, 2)
-    residuals = residuals[:np.product(residuals_shape)].reshape(*residuals_shape)
+    imeas0 = mrcal.measurement_index_boards(0, **optimization_inputs)
+    Nmeas  = mrcal.num_measurements_boards(**optimization_inputs)
+    residuals_board = residuals[imeas0:imeas0+Nmeas].reshape(*residuals_shape)
 
     # shape (Nobservations, object_height_n, object_width_n, 3)
     indices_frame_camera = optimization_inputs['indices_frame_camintrinsics_camextrinsics'][...,:2]
 
     # shape (Nobservations, object_height_n, object_width_n)
-    idx = np.ones( observations.shape[:-1], dtype=bool)
+    idx = np.ones( observations_board.shape[:-1], dtype=bool)
 
-    if i_cam is not None:
+    if icam_intrinsics is not None:
         # select residuals from THIS camera
-        idx[indices_frame_camera[:,1] != i_cam, ...] = False
+        idx[indices_frame_camera[:,1] != icam_intrinsics, ...] = False
 
     # select non-outliers
-    idx[ observations[...,2] <= 0.0 ] = False
+    idx[ observations_board[...,2] <= 0.0 ] = False
 
-    # shape (N,2)
-    return residuals[idx, ...]
+    if not return_observations:
+        # shape (N,2)
+        return \
+            residuals_board   [idx, ...    ]
+    else:
+        # shape (N,2), (N,2)
+        return \
+            residuals_board   [idx, ...    ], \
+            observations_board[idx, ..., :2]
+
+
+def residuals_point(optimization_inputs,
+                    *,
+                    icam_intrinsics     = None,
+                    residuals           = None,
+                    return_observations = False):
+    r'''Compute and return the point observation residuals
+
+SYNOPSIS
+
+    model = mrcal.cameramodel(model_filename)
+
+    gp.plot( mrcal.residuals_point(
+               optimization_inputs = model.optimization_inputs(),
+               icam_intrinsics     = icam_intrinsics ).ravel(),
+             histogram = True,
+             binwidth  = 0.02 )
+
+    ... A plot pops up showing the empirical distribution of point fit
+    ... errors in this solve. For the given camera only
+
+Given a calibration solve, returns the residuals of point observations, throwing
+out outliers and, optionally, selecting the residuals from a specific camera.
+These are the weighted reprojection errors present in the measurement vector.
+
+If no points are present in the solve I return arrays of appropriate shape with
+N = 0
+
+ARGUMENTS
+
+- optimization_inputs: the optimization inputs dict passed into and returned
+  from mrcal.optimize(). This describes the solved optimization problem that
+  we're visualizing
+
+- icam_intrinsics: optional integer to select the camera whose residuals we're
+  visualizing If omitted or None, we report the residuals for ALL the cameras
+  together.
+
+- residuals: optional numpy array of shape (Nmeasurements,) containing the
+  optimization residuals. If omitted or None, this will be recomputed. To use a
+  cached value, pass the result of mrcal.optimize(**optimization_inputs)['x'] or
+  mrcal.optimizer_callback(**optimization_inputs)[1]
+
+- return_observations: optional boolean, defaulting to False. if
+  return_observations: we return a tuple (residuals,observations) instead of
+  just residuals
+
+RETURNED VALUES
+
+if not return_observations:
+
+  we return a numpy array of shape (N,2) of all the residuals. N is the number
+  of pixel observations remaining after outliers and other cameras are thrown
+  out
+
+else:
+
+  we return a tuple:
+
+  - The same residuals array as before
+
+  - The corresponding observation points in a numpy array of shape (N,2). These
+    are a slice of observations_point[] corresponding to each residual
+
+    '''
+
+    # shape (Nobservations, 3)
+    observations_point = optimization_inputs.get('observations_point')
+    if observations_point is None or observations_point.size == 0:
+        # no point observations
+        if not return_observations:
+            # shape (N,2)
+            return \
+                np.zeros((0,2), dtype=float)
+        else:
+            # shape (N,2), (N,2)
+            return \
+                np.zeros((0,2), dtype=float), \
+                np.zeros((0,2), dtype=float)
+
+    if residuals is None:
+        # Flattened residuals. This is ALL the measurements: chessboard, point,
+        # regularization...
+        residuals = \
+            mrcal.optimizer_callback(**optimization_inputs,
+                                     no_jacobian      = True,
+                                     no_factorization = True)[1]
+
+    Nobservations = len(observations_point)
+
+    imeas0 = mrcal.measurement_index_points(0, **optimization_inputs)
+    Nmeas  = mrcal.num_measurements_points(**optimization_inputs)
+
+    # shape (Nobservations, 2)
+    #
+    # Each observation row is (err_x, err_y, range_penalty). I ignore the
+    # range_penalty here
+    residuals_point = residuals[imeas0:imeas0+Nmeas].reshape(Nobservations,3)[:,:2]
+
+    indices_point_camera = optimization_inputs['indices_point_camintrinsics_camextrinsics'][...,:2]
+
+    # shape (Nobservations,)
+    idx = np.ones( (Nobservations,), dtype=bool)
+
+    if icam_intrinsics is not None:
+        # select residuals from THIS camera
+        idx[indices_point_camera[:,1] != icam_intrinsics, ...] = False
+
+    # select non-outliers
+    idx[ observations_point[...,2] <= 0.0 ] = False
+
+    if not return_observations:
+        # shape (N,2)
+        return \
+            residuals_point   [idx, ...    ]
+    else:
+        # shape (N,2), (N,2)
+        return \
+            residuals_point   [idx, ...    ], \
+            observations_point[idx, ..., :2]
