@@ -67,7 +67,7 @@ testutils.confirm_equal( m.extrinsics_rt_fromref(), rt_0r )
 # regions
 m = mrcal.cameramodel(f"{testdir}/data/cam0.opencv8.cameramodel")
 testutils.confirm_equal( m.valid_intrinsics_region(), None,
-                         "read 'valid_intrinsics_region is None' properly")
+                         msg = "read 'valid_intrinsics_region is None' properly")
 
 r_open = np.array(((0,  0),
                    (0, 10),
@@ -82,19 +82,19 @@ r_empty = np.zeros((0,2))
 
 m.valid_intrinsics_region(r_open)
 testutils.confirm_equal( m.valid_intrinsics_region(), r_closed,
-                         "was able to set an open valid_intrinsics_region and to see it be closed")
+                         msg="was able to set an open valid_intrinsics_region and to see it be closed")
 m.write(f'{workdir}/out.cameramodel')
 m1 = mrcal.cameramodel(f'{workdir}/out.cameramodel')
 testutils.confirm_equal( m1.valid_intrinsics_region(), r_closed,
-                         "read 'valid_intrinsics_region' properly")
+                         msg="read 'valid_intrinsics_region' properly")
 
 m.valid_intrinsics_region(r_empty)
 testutils.confirm_equal( m.valid_intrinsics_region(), r_empty,
-                         "was able to set an empty valid_intrinsics_region")
+                         msg="was able to set an empty valid_intrinsics_region")
 m.write(f'{workdir}/out.cameramodel')
 m1 = mrcal.cameramodel(f'{workdir}/out.cameramodel')
 testutils.confirm_equal( m1.valid_intrinsics_region(), r_empty,
-                         "read empty valid_intrinsics_region properly")
+                         msg="read empty valid_intrinsics_region properly")
 
 
 # Make sure we can read model data with extra spacing
@@ -121,7 +121,7 @@ with io.StringIO(string) as f:
     m = mrcal.cameramodel(f)
 
     testutils.confirm_equal( m.intrinsics()[1], [ 1761.181055, 1761.250444, 1965.706996, 1087.518797, -0.01266096516, 0.03590794372, -0.0002547045941, 0.0005275929652, 0.01968883397, 0.01482863541, -0.0562239888, 0.0500223357,],
-                             "extra spaces don't confuse the parser")
+                             msg="extra spaces don't confuse the parser")
 
 
 
