@@ -167,11 +167,11 @@ for kwargs in all_test_kwargs:
     J2 = J.copy()
     mrcal.pack_state(   J2, **optimization_inputs)
     mrcal.unpack_state( J2, **optimization_inputs)
-    testutils.confirm_equal( J2, J, "unpack(pack(J)) = J")
+    testutils.confirm_equal( J2, J, msg="unpack(pack(J)) = J")
     J2 = J.copy()
     mrcal.unpack_state( J2, **optimization_inputs)
     mrcal.pack_state(   J2, **optimization_inputs)
-    testutils.confirm_equal( J2, J, "pack(unpack(J)) = J")
+    testutils.confirm_equal( J2, J, msg="pack(unpack(J)) = J")
 
 
     # I compare full-state J so that I can change SCALE_... without breaking the
