@@ -656,12 +656,11 @@ if not did_sample:
     ( intrinsics_sampled,         \
       extrinsics_sampled_mounted, \
       frames_sampled,             \
+      points_sampled,             \
       calobject_warp_sampled,     \
       optimization_inputs_sampled ) = \
-          calibration_sample( args.Nsamples, args.Ncameras, args.Nframes,
-                              Nintrinsics,
+          calibration_sample( args.Nsamples,
                               optimization_inputs_baseline,
-                              observations_true,
                               args.q_calibration_stdev,
                               fixedframes)
 
